/*
 * This file is part of libphidget22java
 *
 * Copyright (c) 2017-2022 Phidgets Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "phidget_jni.h"

JavaVM *ph_vm = 0;

jclass ph_exception_class;
jmethodID ph_exception_cons1;
jmethodID ph_exception_cons2;

static jclass asyncResult_class;
static jmethodID asyncResult_cons;
static jclass asyncListener_class;
static jmethodID onAsyncCallback_mid;

typedef void(CCONV *mos_task_thread_exit_callback)	(void *ctx);
extern void mos_task_set_thread_exit_handler(mos_task_thread_exit_callback fptr, void *ctx);

static void detachCurrentThreadFromJavaVM(void *ctx)
{
	JNIEnv *env;
	if (ph_vm != NULL && *ph_vm != NULL)
	{
		if ((*ph_vm)->GetEnv(ph_vm, (void **)&env, JNI_VERSION_1_4) != JNI_EDETACHED)
		{
			(*ph_vm)->DetachCurrentThread(ph_vm);
		}
	}
}

jint JNICALL
JNI_OnLoad(JavaVM *vm, void *reserved)
{
	JNIEnv *env;
	jint version = 0;

	ph_vm = vm;

	if((*ph_vm)->GetEnv(ph_vm, (void **)&env, JNI_VERSION_1_4) != JNI_OK)
		return -1;

	if(!(version = (*env)->GetVersion(env)))
		JNI_ABORT_STDERR("Couldn't get version\nJNI Version: 0x%08x", version);

	// PhidgetException is required
	if (!(ph_exception_class = (*env)->FindClass(env, "com/phidget22/PhidgetException")))
		JNI_ABORT_STDERR("Coulnd't FindClass com/phidget22/PhidgetException");
	if (!(ph_exception_class = (jclass)(*env)->NewGlobalRef(env, ph_exception_class)))
		JNI_ABORT_STDERR("Couldn't create global ref ph_exception_class");
	if (!(ph_exception_cons1 = (*env)->GetMethodID(env, ph_exception_class, "<init>", "(ILjava/lang/String;Ljava/lang/String;)V")))
		JNI_ABORT_STDERR("Couldn't get Method ID <init> from ph_exception_class");
	if (!(ph_exception_cons2 = (*env)->GetMethodID(env, ph_exception_class, "<init>", "(ILjava/lang/String;)V")))
		JNI_ABORT_STDERR("Couldn't get Method ID <init> from ph_exception_class");

	// Async support is required
	if (!(asyncResult_class = (*env)->FindClass(env, "com/phidget22/AsyncResult")))
		JNI_ABORT_STDERR("Couldn't FindClass com/phidget22/AsyncResult");
	if (!(asyncResult_class = (jclass)(*env)->NewGlobalRef(env, asyncResult_class)))
		JNI_ABORT_STDERR("Couldn't create global ref AsyncResult_class");
	if (!(asyncResult_cons = (*env)->GetMethodID(env, asyncResult_class, "<init>", "(Lcom/phidget22/Phidget;Lcom/phidget22/ErrorCode;)V")))
		JNI_ABORT_STDERR("Couldn't get method ID <init> from AsyncResult_class");

	if (!(asyncListener_class = (*env)->FindClass(env, "com/phidget22/AsyncListener")))
		JNI_ABORT_STDERR("Couldn't FindClass com/phidget22/AsyncListener");
	if (!(asyncListener_class = (jclass)(*env)->NewGlobalRef(env, asyncListener_class)))
		JNI_ABORT_STDERR("Couldn't create global ref asyncListener_class");
	if (!(onAsyncCallback_mid = (*env)->GetMethodID(env, asyncListener_class, "onAsyncCallback", "(Lcom/phidget22/AsyncResult;)V")))
		JNI_ABORT_STDERR("Couldn't get method ID onAsyncCallback");

	com_phidget22_OnLoad(env);
	phidget_enums_OnLoad(env);
	phidget_structs_OnLoad(env);

	mos_task_set_thread_exit_handler(detachCurrentThreadFromJavaVM, NULL);

#ifdef _ANDROID
	if(com_phidget22_usb_Phidget_OnLoad(env) && com_phidget22_usb_Manager_OnLoad(env))
		AndroidUsbJarAvailable = PTRUE;
	else
		AndroidUsbJarAvailable = PFALSE;
#endif

	return JNI_VERSION_1_4;
}

jlong
updateGlobalRef(JNIEnv *env, jobject obj, jfieldID fid, jboolean b)
{
	/*
	 * Manages the global reference held by phidget22 to the handler
	 * target.
	 */
	jlong gr;

	if ((gr = (*env)->GetLongField(env, obj, fid)) != 0)
		(*env)->DeleteGlobalRef(env, (jobject)(uintptr_t)gr);
	gr = b ? (jlong)(uintptr_t)(*env)->NewGlobalRef(env, obj) : 0;
	(*env)->SetLongField(env, obj, fid, gr);

	return gr;
}

void asyncCallbackHandler(PhidgetHandle phid, void *arg, PhidgetReturnCode returnCode) {
	JNIEnv *env;
	jobject asyncResult;
	jobject listener;
	jobject phidget;
	jobjectArray arr;
	jobject errorCodeEnum;

	if ((*ph_vm)->AttachCurrentThreadAsDaemon(ph_vm, (JNIEnvPtr)&env, NULL))
		JNI_ABORT_STDERR("Couldn't AttachCurrentThread");

	arr = (jobjectArray)arg;
	phidget = (*env)->GetObjectArrayElement(env, arr, 0);
	listener = (*env)->GetObjectArrayElement(env, arr, 1);
	errorCodeEnum = (*env)->CallStaticObjectMethod(env, ErrorCode_class, ErrorCode_getEnum_mid, returnCode);

	if (!(asyncResult = (*env)->NewObject(env, asyncResult_class, asyncResult_cons, phidget, errorCodeEnum)))
		JNI_ABORT_STDERR("Couldn't create new AsyncResult object");

	(*env)->CallVoidMethod(env, listener, onAsyncCallback_mid, asyncResult);

	(*env)->DeleteLocalRef(env, errorCodeEnum);
	(*env)->DeleteLocalRef(env, asyncResult);
	(*env)->DeleteLocalRef(env, phidget);
	(*env)->DeleteLocalRef(env, listener);
	(*env)->DeleteGlobalRef(env, arr);
}

