/*
 * This file is part of libphidget22java
 *
 * Copyright (c) 2017-2022 Phidgets Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef PHIDGET_JNI_H
#define PHIDGET_JNI_H

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>

#include "jni.h"

#include "phidget22.h"
#include "phidget_enums_jni.h"

 /* Define here structs which are not in the library */
typedef struct {
	const char *code;
	int bitCount;
} IRCode, *IRCodeHandle;

typedef struct {
	const char *code;
	PhidgetIR_CodeInfo codeInfo;
} IRLearnedCode, *IRLearnedCodeHandle;

typedef struct {
	const char *tagString;
	PhidgetRFID_Protocol protocol;
} RFIDTag, *RFIDTagHandle;

typedef struct {
	uint64_t size;
	int keepCount;
} LogRotating, *LogRotatingHandle;

typedef struct {
	uint32_t distances[8];
	uint32_t amplitudes[8];
	uint32_t count;
} DistanceSensorSonarReflections, *DistanceSensorSonarReflectionsHandle;

#include "phidget_structs_jni.h"

extern const char * CCONV Phidget_strerror(PhidgetReturnCode error);
extern void asyncCallbackHandler(PhidgetHandle phid, void *ctx, PhidgetReturnCode returnCode);

void CCONV setOnJavaDetachCurrentThreadHandler(void(CCONV *fptr)(void));

extern void com_phidget22_OnLoad(JNIEnv *env);

#ifdef _ANDROID
#define JNIEnvPtr const struct JNINativeInterface ***
#else
#define JNIEnvPtr void **
#endif

extern JavaVM *ph_vm;

extern jclass ph_exception_class;
extern jmethodID ph_exception_cons1;
extern jmethodID ph_exception_cons2;
extern jclass phid_class;
extern jmethodID phid_cons;
extern jfieldID phidget_handle_fid;

#ifdef _ANDROID
extern int AndroidUsbJarAvailable;
int com_phidget22_usb_Manager_OnLoad(JNIEnv *);
int com_phidget22_usb_Phidget_OnLoad(JNIEnv *);
#endif

jlong updateGlobalRef(JNIEnv *env, jobject obj, jfieldID fid, jboolean b);

#define JNI_ABORT_STDERR(...) \
do { \
	fprintf(stderr, "\nAn error was encountered in the phidget22java jni library. This program will now be aborted.\n"); \
	fprintf(stderr, "Make sure phidget22.jar and phidget22 C libraries are in sync.\n\n"); \
	fprintf(stderr, "Error Details:\n"); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n\n"); \
	(*env)->ExceptionDescribe(env); \
	(*env)->ExceptionClear(env); \
	abort(); \
} while(0)

#define PH_THROW_LAST_ERR(errno) do { \
	jobject eobj; \
	jstring edesc; \
	jstring edetail; \
	const char *errorDesc; \
	char *errorDetail; \
	size_t errorDetailLen; \
	PhidgetReturnCode code; \
 \
	if(Phidget_getLastError(&code, &errorDesc, NULL, &errorDetailLen) == EPHIDGET_OK && code == errno) { \
		errorDetail = malloc(errorDetailLen); \
		Phidget_getLastError(&code, &errorDesc, errorDetail, &errorDetailLen); \
		if (!(edesc = (*env)->NewStringUTF(env, errorDesc))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(edetail = (*env)->NewStringUTF(env, errorDetail))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons1, code, edesc, edetail))) \
			JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
		(*env)->DeleteLocalRef (env, edetail); \
		free(errorDetail); \
	} else { \
		if (!(edesc = (*env)->NewStringUTF(env, Phidget_strerror(errno)))) \
			JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
		if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons2, errno, edesc))) \
			JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
	} \
	(*env)->DeleteLocalRef (env, edesc); \
	(*env)->Throw(env, (jthrowable)eobj); \
	(*env)->DeleteLocalRef (env, eobj); \
} while(0)

#define PH_THROW(errno) do { \
	jobject eobj; \
	jstring edesc; \
 \
	if (!(edesc = (*env)->NewStringUTF(env, Phidget_strerror(errno)))) \
		JNI_ABORT_STDERR("Couldn't get NewStringUTF"); \
	if (!(eobj = (*env)->NewObject(env, ph_exception_class, ph_exception_cons2, errno, edesc))) \
		JNI_ABORT_STDERR("Couldn't get NewObject ph_exception_class"); \
	(*env)->DeleteLocalRef (env, edesc); \
	(*env)->Throw(env, (jthrowable)eobj); \
	(*env)->DeleteLocalRef (env, eobj); \
} while(0)


#endif
